@@ -12,6 +12,7 @@ use std::sync::{Arc, Mutex};
 use std::thread;
 
 mod git;
+mod log_utils;
 pub mod lint_config;
 pub mod lint_message;
 pub mod linter;
@@ -19,7 +20,8 @@ pub mod path;
 pub mod render;
 
 use git::get_changed_files;
-use git::get_paths_cmd_files;
+use git::get_git_root;
+use git::get_paths_from_cmd;
 use lint_message::LintMessage;
 use render::PrintedLintErrors;
 
@@ -82,14 +84,30 @@ fn get_paths_from_input(paths: Vec<String>) -> Result<Vec<AbsPath>> {
     let mut ret = Vec::new();
     for path in &paths {
         let path = AbsPath::new(PathBuf::from(path))
-            .with_context(|| format!("Failed to lint provided file: '{}'", path))?;
+            .with_context(|| format!("Failed to find provided file: '{}'", path))?;
         ret.push(path);
     }
     Ok(ret)
 }
 
+fn get_paths_from_file(file: AbsPath) -> Result<Vec<AbsPath>> {
+    let file = std::fs::read_to_string(file.as_pathbuf()).with_context(|| {
+        format!(
+            "Failed to read file specified in `--paths-from`: '{}'",
+            file.as_pathbuf().display()
+        )
+    })?;
+    let files = file
+        .trim()
+        .lines()
+        .map(|l| l.to_string())
+        .collect::<Vec<String>>();
+    get_paths_from_input(files)
+}
+
 pub enum PathsToLint {
     Auto,
+    PathsFile(AbsPath),
     PathsCmd(String),
     Paths(Vec<String>),
 }
@@ -106,20 +124,30 @@ pub fn do_lint(
         linters.iter().map(|l| &l.code).collect::<Vec<_>>()
     );
 
-    // Too lazy to learn rust's fancy concurrent programming stuff, just spawn a thread per linter and join them.
-    let all_lints = Arc::new(Mutex::new(HashMap::new()));
-    let files = match paths_to_lint {
-        PathsToLint::Auto => get_changed_files()?,
-        PathsToLint::PathsCmd(paths_cmd) => get_paths_cmd_files(paths_cmd)?,
+    let mut files = match paths_to_lint {
+        PathsToLint::Auto => {
+            let git_root = get_git_root()?;
+            get_changed_files(git_root)?
+        }
+        PathsToLint::PathsCmd(paths_cmd) => get_paths_from_cmd(paths_cmd)?,
         PathsToLint::Paths(paths) => get_paths_from_input(paths)?,
+        PathsToLint::PathsFile(file) => get_paths_from_file(file)?,
     };
+
+    // Sort and unique the files so we pass a consistent ordering to linters
+    files.sort();
+    files.dedup();
+
     let files = Arc::new(files);
 
-    debug!("Linting files: {:#?}", files);
+    log_utils::log_files("Linting files: ", &files);
 
     let mut thread_handles = Vec::new();
     let spinners = Arc::new(MultiProgress::new());
 
+    // Too lazy to learn rust's fancy concurrent programming stuff, just spawn a thread per linter and join them.
+    let all_lints = Arc::new(Mutex::new(HashMap::new()));
+
     for linter in linters {
         let all_lints = Arc::clone(&all_lints);
         let files = Arc::clone(&files);
@@ -189,3 +217,32 @@ pub fn do_lint(
         PrintedLintErrors::Yes => Ok(1),
     }
 }
+
+#[cfg(test)]
+mod tests {
+    use super::*;
+    use std::io::Write;
+    use tempfile::NamedTempFile;
+
+    #[test]
+    fn test_paths_file() -> Result<()> {
+        let file1 = NamedTempFile::new()?;
+        let file2 = NamedTempFile::new()?;
+
+        let mut paths_file = NamedTempFile::new()?;
+
+        writeln!(paths_file, "{}", file1.path().display())?;
+        writeln!(paths_file, "{}", file2.path().display())?;
+
+        let paths_file = AbsPath::new(paths_file.path().to_path_buf())?;
+        let paths = get_paths_from_file(paths_file)?;
+
+        let file1_abspath = AbsPath::new(file1.path().to_path_buf())?;
+        let file2_abspath = AbsPath::new(file2.path().to_path_buf())?;
+
+        assert!(paths.contains(&file1_abspath));
+        assert!(paths.contains(&file2_abspath));
+
+        Ok(())
+    }
+}
