@@ -3,7 +3,7 @@ use std::io::Write;
 use std::path::PathBuf;
 use std::{cmp, collections::HashMap, fs};
 
-use anyhow::{Context, Result};
+use anyhow::{anyhow, Result};
 use console::{style, Style, Term};
 use similar::{ChangeTag, DiffableStr, TextDiff};
 use textwrap::indent;
@@ -23,11 +23,9 @@ pub fn render_lint_messages_json(
     lint_messages: &HashMap<Option<String>, Vec<LintMessage>>,
 ) -> Result<PrintedLintErrors> {
     let mut printed = false;
-    for (_, lint_message) in lint_messages {
-        for lint_message in lint_message {
-            printed = true;
-            writeln!(stdout, "{}", serde_json::to_string(lint_message)?)?;
-        }
+    for lint_message in lint_messages.values().flatten() {
+        printed = true;
+        writeln!(stdout, "{}", serde_json::to_string(lint_message)?)?;
     }
 
     if printed {
@@ -42,7 +40,7 @@ pub fn render_lint_messages(
     lint_messages: &HashMap<Option<String>, Vec<LintMessage>>,
 ) -> Result<PrintedLintErrors> {
     if lint_messages.is_empty() {
-        writeln!(stdout, "{} {}", style("ok").green(), "No lint issues.")?;
+        writeln!(stdout, "{} No lint issues.", style("ok").green())?;
 
         return Ok(PrintedLintErrors::No);
     }
@@ -117,30 +115,59 @@ pub fn render_lint_messages(
 }
 
 // Write formatted context lines, with an styled indicator for which line the lint is about
-fn write_context(stdout: &mut impl Write, path: &String, highlight_line: &usize) -> Result<()> {
+fn write_context(stdout: &mut impl Write, path: &str, highlight_line: &usize) -> Result<()> {
     stdout.write_all(b"\n")?;
-    let file = fs::read_to_string(path).context(format!(
-        "Error reading file: '{}' when rendering lints",
-        path
-    ))?;
-    let lines = file.tokenize_lines();
-    let line_idx = highlight_line.saturating_sub(1);
-    let max_idx = lines.len().saturating_sub(1);
-    let start_idx = line_idx.saturating_sub(CONTEXT_LINES);
-    let end_idx = cmp::min(max_idx, line_idx + CONTEXT_LINES);
-    for cur_idx in start_idx..=end_idx {
-        let line = lines
-            .get(cur_idx)
-            .ok_or(anyhow::Error::msg("TODO line mismatch"))?;
-        let line_number = cur_idx + 1;
+    let file = fs::read_to_string(path);
+    match file {
+        Ok(file) => {
+            let lines = file.tokenize_lines();
 
-        // Wrlte `123 |  my failing line content
+            let highlight_idx = highlight_line.saturating_sub(1);
 
-        if cur_idx == line_idx {
-            // Highlight the actually failing line with a chevron + different color
-            write!(stdout, "    >>> {}  |{}", style(line_number).dim(), style(line).yellow())?;
-        } else {
-            write!(stdout, "        {}  |{}", style(line_number).dim(), line)?;
+            let max_idx = lines.len().saturating_sub(1);
+            let start_idx = highlight_idx.saturating_sub(CONTEXT_LINES);
+            let end_idx = cmp::min(max_idx, highlight_idx + CONTEXT_LINES);
+
+            for cur_idx in start_idx..=end_idx {
+                let line = lines
+                    .get(cur_idx)
+                    .ok_or_else(|| anyhow!("TODO line mismatch"))?;
+                let line_number = cur_idx + 1;
+
+                let max_line_number = max_idx + 1;
+                let max_pad = max_line_number.to_string().len();
+
+                // Write `123 |  my failing line content
+                if cur_idx == highlight_idx {
+                    // Highlight the actually failing line with a chevron + different color
+                    write!(
+                        stdout,
+                        "    >>> {:>width$}  |{}",
+                        style(line_number).dim(),
+                        style(line).yellow(),
+                        width = max_pad
+                    )?;
+                } else {
+                    write!(
+                        stdout,
+                        "        {:>width$}  |{}",
+                        style(line_number).dim(),
+                        line,
+                        width = max_pad
+                    )?;
+                }
+            }
+        }
+        Err(e) => {
+            let msg = textwrap::indent(
+                &format!(
+                    "Could not retrieve source context: {}\n\
+                    This is typically a linter bug.",
+                    e
+                ),
+                spaces(8),
+            );
+            write!(stdout, "{}", style(msg).red())?;
         }
     }
     stdout.write_all(b"\n")?;
@@ -148,11 +175,7 @@ fn write_context(stdout: &mut impl Write, path: &String, highlight_line: &usize)
 }
 
 // Write the context, computing and styling a diff from the original to the suggested replacement.
-fn write_context_diff(
-    stdout: &mut impl Write,
-    original: &String,
-    replacement: &String,
-) -> Result<()> {
+fn write_context_diff(stdout: &mut impl Write, original: &str, replacement: &str) -> Result<()> {
     writeln!(
         stdout,
         "\n    {}",
@@ -160,6 +183,20 @@ fn write_context_diff(
     )?;
     stdout.write_all(b"\n")?;
     let diff = TextDiff::from_lines(original, replacement);
+
+    let mut max_line_number = 1;
+    for (_, group) in diff.grouped_ops(3).iter().enumerate() {
+        for op in group {
+            for change in diff.iter_inline_changes(op) {
+                let old_line = change.old_index().unwrap_or(0) + 1;
+                let new_line = change.new_index().unwrap_or(0) + 1;
+                max_line_number = cmp::max(max_line_number, old_line);
+                max_line_number = cmp::max(max_line_number, new_line);
+            }
+        }
+    }
+    let max_pad = max_line_number.to_string().len();
+
     for (idx, group) in diff.grouped_ops(3).iter().enumerate() {
         if idx > 0 {
             writeln!(stdout, "{:-^1$}", "-", 80)?;
@@ -172,6 +209,7 @@ fn write_context_diff(
                     ChangeTag::Equal => (" ", Style::new().dim()),
                 };
                 let changeset = Changeset {
+                    max_pad,
                     old: change.old_index(),
                     new: change.new_index(),
                 };
@@ -228,6 +266,8 @@ fn spaces(len: u8) -> &'static str {
 }
 
 struct Changeset {
+    // The length of the largest line number we'll be printing.
+    max_pad: usize,
     old: Option<usize>,
     new: Option<usize>,
 }
@@ -241,21 +281,23 @@ impl fmt::Display for Changeset {
                 // +1 because we want to print the line number, not the vector index.
                 let old = old + 1;
                 let new = new + 1;
-                write!(f, "{}  {}", old, new)
+                write!(
+                    f,
+                    "{:>left_pad$}  {:>right_pad$}",
+                    old,
+                    new,
+                    left_pad = self.max_pad,
+                    right_pad = self.max_pad,
+                )
             }
             // In cases where old/new are missing, do an approximation:
             // '1234      '
             //        ^^^^ length of '1234' mirrored to the other side
             //      ^^ two spaces still
-            (Some(old), None) => {
-                let old = old + 1;
-                let total_length = old.to_string().len() * 2 + 2;
-                write!(f, "{:<width$}", old, width = total_length)
-            }
+            (Some(old), None) => write!(f, "{:>width$}  {:width$}", old, " ", width = self.max_pad),
             (None, Some(new)) => {
                 let new = new + 1;
-                let total_length = new.to_string().len() * 2 + 2;
-                write!(f, "{:>width$}", new, width = total_length)
+                write!(f, "{:width$}  {:>width$}", " ", new, width = self.max_pad)
             }
             (None, None) => unreachable!(),
         }
